package net.sf.RecordEditor.diff;

import java.io.PrintStream;

import net.sf.RecordEditor.re.layoutSelection.LayoutSelectionDB;
import net.sf.RecordEditor.re.layoutSelection.LayoutSelectionDBCreator;
import net.sf.RecordEditor.utils.CopyBookDbReader;
import net.sf.RecordEditor.utils.common.Common;
import net.sf.RecordEditor.xml.classes.DiffDefinition;
import net.sf.RecordEditor.xml.classes.ReadXml;
import utilityClasses.ParseArguments;

public class BatchDiff {

	private static final String FILE1 = "-file1";
	private static final String FILE2 = "-file2";
	private static final String SCHEMA1 = "-schema1";
	private static final String SCHEMA2 = "-schema2";
	private static final String SINGLE_TABLE = "-singleTable";
	private static final String XML = "-xml";
	private static final String HTML_FILE = "-htmlFile";

	private static String[] validArgs = {
			"-h", "-help", "-?", FILE1, FILE2, SCHEMA1, SCHEMA2, SINGLE_TABLE, XML, HTML_FILE
	};
	
	private final String file1, file2, schema, schema1, schema2, xml, htmlFile;
	
	private final boolean isOk, help, singleTable;
	
	private BatchDiff(String[] args) {
		ParseArguments pa = new ParseArguments(validArgs, args);
		String singleTableStr = pa.getArg(SINGLE_TABLE);
		
		help = pa.getArg("-?", pa.get2Args("-h", "-help", null)) != null;
		file1 = pa.getArg(FILE1, "");
		file2 = pa.getArg(FILE2, "");
		schema1 = pa.getArg(SCHEMA1, "");
		schema2 = pa.getArg(SCHEMA2, "");
		
		schema = schema1.length() == 0 ? schema2 : schema1;
		singleTable = singleTableStr != null && (singleTableStr.length() == 0 || "yes".equalsIgnoreCase(singleTableStr));
		xml = pa.getArg(XML, "");
		htmlFile = pa.getArg(HTML_FILE, "");
		
		
		StringBuilder msg = new StringBuilder();
		
		if (help) {
			printHelp(System.out);
			isOk = false;
		} else {
			if (xml.length() == 0) {
				if (file1.length() == 0 || file2.length() == 0 ) {
					msg.append("You must supply either a Xml definition or 2 file Names (-file1/2)\n");
				}
			
				if (schema.length() == 0) {
					msg.append("You must supply either a Xml definition or a schema (layout) definition (schema1/2)\n");
				}
			}
		
			if (htmlFile.length() == 0) {
				msg.append("You must supply output html file (-htmlFile) \n");
			}
			
			isOk = msg.length() == 0;
			if (! isOk) {
				printHelp(System.err);
				System.err.print("Errors:\n\n" + msg);
			}
		}
	}
	
	
	private DiffDefinition getDiffDefinition() throws Exception {
		DiffDefinition def;
		
		if (xml.length() > 0) {
			def = new ReadXml(xml).readDiffDefinition();
		} else {
			def = new DiffDefinition();
			if (schema1.equalsIgnoreCase(schema2)  || schema1.length() == 0 ||  schema2.length() == 0 ) {
				def.type = DiffDefinition.TYPE_SINGLE_LAYOUT;
				def.getLayoutDetails().name = schema;
			} else {
				def.type = DiffDefinition.TYPE_TWO_LAYOUTS;
				def.getNewfile().getLayoutDetails().name = schema1.length() > 0 ? schema1 : schema;
				def.getOldfile().getLayoutDetails().name = schema2.length() > 0 ? schema2 : schema;
			}
		}
		def.getHtmlOptions().htmlFile = htmlFile;
		def.getHtmlOptions().singleTable = singleTable;
		
		if (file1.length() > 0) {
			def.getNewfile().name = file1;
		}
		if (file2.length() > 0) {
			def.getOldfile().name = file2;
		}
		return def;
	}
	
	
	private void printHelp(PrintStream out) {
		
		out.println();
		out.println("Parameters:");
		out.println("\t" + FILE1 + "/2\t- files to be compared" );
		out.println("\t" + SCHEMA1 + "/2\t- schema (layout) of the files to be compared to be compared" );
		out.println("\t" + SINGLE_TABLE + "\t- Display as one Table (or a table per row" );
		out.println("\t" + XML + "\t- saved compare definition" );
		out.println("\t" + HTML_FILE + "\t- Output HTML file" );
		out.println();
		out.println("Typically you would either supply a saved diff-definition:");
		out.println("\t java -jar run.jar net.sf.RecordEditor.diff.BatchDiff -xml SavedDiff.xml -outHtml out.html");
		out.println("or the files and schema (layout)");
		out.println("\t java -jar run.jar net.sf.RecordEditor.diff.Batch diff -file1 file1.txt -schema1 MyLayout -file2 file2.txt -outHtml out.html");
		out.println();
	}
	

	
	public static void main(String[] args) throws Exception {
		BatchDiff bd = new BatchDiff(args);
		
		if (bd.isOk) {
			LayoutSelectionDBCreator layoutSelectionDBCreator = new LayoutSelectionDBCreator(CopyBookDbReader.getInstance());
			LayoutSelectionDB layoutReader1 = layoutSelectionDBCreator.create();
			LayoutSelectionDB layoutReader2 = layoutSelectionDBCreator.create();
			
			int connectionIndex = Common.getConnectionIndex();
			layoutReader1.setDatabaseIdx(connectionIndex);
			layoutReader2.setDatabaseIdx(connectionIndex);
			
			DoCompare.getInstance().writeHtml(layoutReader1, layoutReader2, bd.getDiffDefinition());
		}
				
	}
}
