package net.sf.RecordEditor.re.util.csv;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Pattern;

import net.sf.JRecord.Common.Conversion;
import net.sf.JRecord.External.CopybookLoaderFactory;
import net.sf.JRecord.Numeric.ICopybookDialects;
import net.sf.cb2xml.analysis.Item;
import net.sf.cb2xml.def.Cb2xmlConstants;

public class SchemaAnalyser {

	private static final int READ_LIMIT = 320000;
	private static final String COBOL_REGEX = ".*\\s*[0-9]{1,2}\\s+[a-z0-9\\-_]+\\s+pic\\s+[xa9\\-\\+s].*";
//	private static final String COBOL_REGEX = ".*\\s*[0-9]{1,2}\\s+[a-z0-9\\-_]+.*+pic\\s*+[xa9\\-\\+\\s].*";
	private static Pattern cobolPattern;
	private static final int[] COBOL_FORMATS = { 
			Cb2xmlConstants.USE_LONG_LINE,
			Cb2xmlConstants.USE_COLS_6_TO_80,
			Cb2xmlConstants.USE_STANDARD_COLUMNS,
			Cb2xmlConstants.FREE_FORMAT
	};
	
	
	private final String filename;
	private final int typeOfSchema;
	private  int cobolCopybookFormat = -1;
//	private final String copyBookData;
	
	

	public SchemaAnalyser(String filename) {
		this(filename, null);
	}

	public SchemaAnalyser(String filename, String cpyData) {
		super();
		this.filename = filename;
		this.typeOfSchema = determineSchemaType(cpyData);
	}


	public int schemaType() {
		return typeOfSchema;
	}

	private int determineSchemaType(String copyBookData) {
		String lcFilename = filename.toLowerCase();
		int schemaType = -1;
		
		if (lcFilename.endsWith(".cbl") || lcFilename.endsWith(".cob")) {
			schemaType = CopybookLoaderFactory.COBOL_LOADER;
		} else if (lcFilename.endsWith(".xml")) {
			schemaType = CopybookLoaderFactory.RECORD_EDITOR_XML_LOADER;
		} else if (lcFilename.endsWith(".tsv") || lcFilename.endsWith(".retsv")) {
			schemaType = CopybookLoaderFactory.TAB_NAMES_1ST_LINE_LOADER;
		} else if (lcFilename.endsWith(".csv")) {
			schemaType = CopybookLoaderFactory.COMMA_NAMES_1ST_LINE_LOADER;
		}
		
		if (copyBookData != null) {
			return checkCobolCopybook(schemaType, copyBookData, true);
		}
		int pos = 0;
		FileReader f = null;
		char[] chars = new char[READ_LIMIT];
		try {
			f = new FileReader(filename);
			
			int num = f.read(chars);
			while (num > 0 && (pos += num) < chars.length) {
				num = f.read(chars, pos, chars.length - pos);
			}
			f.close();
		} catch (FileNotFoundException e) {
			
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (f != null) {
				try {
					f.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		
		if (pos > 0) {
			schemaType = checkCobolCopybook(schemaType, new String(chars, 0, pos), pos >= chars.length);
		}
		
		return schemaType;
	}

	private int checkCobolCopybook(int schemaType, String s, boolean readLimitreached) {
		schemaType = checkSchemaText(schemaType, s);
		
		if (schemaType == CopybookLoaderFactory.COBOL_LOADER) {
			this.cobolCopybookFormat = checkCobolCopybookFormat(s, readLimitreached);
		}
		return schemaType;
	}

	public static int checkSchemaText(int schemaType, String s) {
		String lc = s.toLowerCase();
		String sTrim = lc.trim();
		if (sTrim.startsWith("<")) {
			if (sTrim.startsWith("<?")) {
				int idx = sTrim.indexOf("<", 2);
				sTrim = sTrim.substring(idx);
			}
//			System.out.println("b)  *** " + sTrim.substring(0, Math.min(100, sTrim.length()))
//					+ " ~ " + CopybookLoaderFactory.RECORD_EDITOR_XML_LOADER);
			if (sTrim.startsWith("<record ")) {
				return CopybookLoaderFactory.RECORD_EDITOR_XML_LOADER;
			}
//			System.out.println("c)  *** " + sTrim.substring(0, Math.min(100, sTrim.length())));
			if (sTrim.startsWith("<copybook ")) {
				return CopybookLoaderFactory.CB2XML_LOADER;
			}
		}
		
		if (cobolPattern == null) {
			cobolPattern = Pattern.compile(COBOL_REGEX);
		}
		if (cobolPattern.matcher(	
					Conversion.replace(
						Conversion.replace(lc, "\r", " "), "\n", " "))
				.matches()) {
			return CopybookLoaderFactory.COBOL_LOADER;
		}
		
		int commaCount = 0;
		int tabCount   = 0;
		int lineCount  = 0;
		int lineIdx    = -1;
		int idx;
		
		do {
			int lineStart = lineIdx + 1;
			lineIdx = s.indexOf('\n', lineStart);
			idx = s.indexOf('\t', lineStart);
			if (idx >= 0 && idx < lineIdx) {
				tabCount += 1;
			}
			idx = s.indexOf(',', lineStart);
			if (idx >= 0 && idx  < lineIdx) {
				commaCount += 1;
			}
			lineCount += 1;
		} while (lineCount < 100 && lineIdx >= 0);
		
		if (lineIdx < 0) {
			lineCount -= 1;
		}
		
		if (lineCount > 1) {
			if (tabCount * 10 / lineCount > 8) {
				return CopybookLoaderFactory.TAB_NAMES_1ST_LINE_LOADER;
			}
			if (commaCount * 10 / lineCount > 8) {
				return CopybookLoaderFactory.COMMA_NAMES_1ST_LINE_LOADER;
			}
		}
		return schemaType;
	}
	
	private static int checkCobolCopybookFormat(String s, boolean readLimitreached) {
		int cpyFormat = Cb2xmlConstants.USE_STANDARD_COLUMNS;
		
		int startNums   = 0;
		int lineEndNums = 0;
		int maxLineLength = 0;
//		int freeFormat = 0;
//		int freeFormatNums = 0;
		boolean[] test = new boolean[Cb2xmlConstants.FREE_FORMAT + 1];
		Arrays.fill(test, true);

		String line;
		try {
			BufferedReader reader = new BufferedReader(new StringReader(s));
			while ((line = reader.readLine()) != null) {
				int length = line.length();
				String line6;
				if (length >= 6) {
					line6 = line.substring(0, 6);

					startNums   += isNum(line6) ? 1 : 0;
				}
				if (! line.trim().startsWith("*")) {
					if (line.endsWith(" ")) {
						int pos = length - 2;
						while (pos >= 0 && line.charAt(pos) == ' ') {
							pos -= 1;
						}
						length = pos + 1;
						line = line.substring(0, length);
					}
					if (length >= 6) {
						lineEndNums += length == 80 && isNum(line.substring(73)) ? 1 : 0;
	//					String trimLine6 = line6.trim() + "   ";
	//					freeFormatNums += trimLine6.charAt(2) == ' ' && isNum(trimLine6.substring(0, 2)) ? 1 : 0;
					}
	//				freeFormat  += s.trim().startsWith("*>") ? 1 : 0;
					maxLineLength = Math.max(maxLineLength, length);
				}
			}
			reader.close();
			test[Cb2xmlConstants.USE_LONG_LINE] = maxLineLength > 80 && lineEndNums < 3;
		} catch (IOException e1) {
		}
		
		test[Cb2xmlConstants.USE_COLS_6_TO_80] = lineEndNums < 3 && maxLineLength > 72; 
		test[Cb2xmlConstants.FREE_FORMAT] = startNums < 3;
		
		int itemCount = 0;
		for (int copybookFormat : COBOL_FORMATS) {
			try {
				if (test[copybookFormat]) {
					net.sf.cb2xml.analysis.Copybook cpy 
						= net.sf.JRecord.External.Def.Cb2Xml.getCopybook(
							new StringReader(s), "Cpy",
							ICopybookDialects.FMT_MAINFRAME, false, 
							copybookFormat, Cb2xmlConstants.CALCULATE_THREAD_SIZE);
					int itmCount = countFields(cpy.getChildItems());
					if (itmCount > itemCount) {
						cpyFormat = copybookFormat;
						itemCount = itmCount;
					}
				}
			} catch (Exception e) {
				System.out.print("\t" + copybookFormat);
			}
		}

		return cpyFormat;
	}
	
	private static boolean isNum(String s) {
		if (s == null || s.length() == 0) { return false; }
		
		for (int i = 0; i < s.length(); i++) {
			char c = s.charAt(i);
			if (c <'0' || c > 9) { return false; }
		}
		return true;
	}
	
	private static int countFields(List<? extends Item> flds) {
		int ret = 0;
		if (flds != null && flds.size() > 0) {
			ret = flds.size();
			for (Item itm : flds) {
				ret += countFields(itm.getChildItems());
			}
		}
		return ret;
	}

	public int cobolCopybookFormat() {
		return cobolCopybookFormat;
	}
	
	
//	public static void main (String[] a) {
//		String regex = ".*\\s*[0-9]{1,2}\\s+[a-zA-Z0-9\\-_]+\\s+pic\\s+[xa9\\-\\+s].*";
//		System.out.println(
//				Pattern.matches(regex, "  pic x(12)")
//				+ " " + Pattern.matches(regex, "  pic 9(12)")
//				+ " " + Pattern.matches(regex, "  picx 9(12)")
//				+ " ! "
//				+ " " + Pattern.matches(regex, " 1 as-d pic x(12)")
//				+ " " + Pattern.matches(regex, " 1 as-d pic 9(12)")
//				+ " " + Pattern.matches(regex, " 1  as-d pic 9(12)")
//				+ " " + Pattern.matches(regex, " 1 as-d  pic 9(12)")
//				+ " " + Pattern.matches(regex, " 1\t as-d     pic 9(12)")
//				+ " " + Pattern.matches(regex, " 1 as-d picx 9(12)")
//				+ " ! "
//				+ " " + Pattern.matches(regex, " 10 as-d pic x(12)")
//				+ " " + Pattern.matches(regex, " 10 as-d pic 9(12)")
//				+ " " + Pattern.matches(regex, " 10  as-d pic 9(12)")
//				+ " " + Pattern.matches(regex, " 10 as-d  pic 9(12)")
//				+ " " + Pattern.matches(regex, " 10\t as-d     pic 9(12)")
//				+ " " + Pattern.matches(regex, " 10 as-d picx 9(12)")
//				+ " ! "
//				+ " " + Pattern.matches(regex, " 01 as-d pic x(12)")
//				+ " " + Pattern.matches(regex, " 01 as-d  pic 9(12)")
//				+ " " + Pattern.matches(regex, " 01 as-d  picx 9(12)")
//			
//				);
//		
//		regex = ".*\\s*[0-9]{1,2}\\s+[a-zA-Z0-9\\-_]+\\sPIC\\s+[XA9\\-\\+s].*";
//		regex = ".*\\sPIC\\s.*";
//		
//		System.out.println(
//					Pattern.matches(regex, " PIC ") + " " +
//					Pattern.matches(regex, "  PIC  ") + " " +
//					Pattern.matches(regex, "x PIC x") + " " +
//					Pattern.matches(regex, " x PIC x") + " " +
//					Pattern.matches(regex, "                05 RD430-PROD-NO                 PIC 9(14).") + " " +
//					Pattern.matches(regex, "                05 RD430-PROD-NO                 PIC 9(14).\n") + " ");
//		System.out.println(
//					Pattern.matches(regex,
//						Conversion.replace(
//						"000900        03  DTAR020-KCODE-STORE-KEY. \n" +                                     
//						"001000            05 DTAR020-KEYCODE-NO      PIC X(08).\n" +                       
//						"001100            05 DTAR020-STORE-NO        PIC S9(03)   COMP-3.  \n" +          
//						"001200        03  DTAR020-DATE               PIC S9(07)   COMP-3.  \n", "\n", " ")             
//					) + " " +
//					Pattern.matches(regex,
//                     Conversion.replace(
//							"000900    01  DTAR020. \n" +                                     
//							"000900        03  DTAR020-KCODE-STORE-KEY. \n" +                                     
//							"001000            05 DTAR020-KEYCODE-NO      PIC X(08).\n" +                       
//							"001100            05 DTAR020-STORE-NO        PIC S9(03)   COMP-3.  \n" +          
//							"001200        03  DTAR020-DATE               PIC S9(07)   COMP-3.  \n", "\n", " ")             
//					)  + " " +
//
//					Pattern.matches(regex,
//	                    Conversion.replace(
//						"000900    01  DTAR020. \n" +                                     
//								"001100            05 DTAR020-STORE-NO        PIC S9(03)   COMP-3.  \n" +          
//								"001200        03  DTAR020-DATE               PIC S9(07)   COMP-3.  \n" , "\n", " ")            
//					)  + " " +
//					Pattern.matches(regex,
//  Conversion.replace(
//							"        01  RD430-PROD-RECORD.\n" + 
//							"            03  RD430-PROD-NO-X.\n" +                                      
//							"                05 RD430-PROD-NO                 PIC 9(14).\n" +       
//							"            03  RD430-PROD-BRAND-ID              PIC X(3).\n" +            
//							"                88 RD430-PROD-BRAND-COLES        VALUE 'TAR'.\n"  , "\n", " ")           
//						)  + " " +
//					Pattern.matches(regex,
//  Conversion.replace(
//							"                05 RD430-PROD-NO                 PIC 9(14).\n" +       
//							"            03  RD430-PROD-BRAND-ID              PIC X(3).\n" +            
//							"                88 RD430-PROD-BRAND-COLES        VALUE 'TAR'.\n", "\n", " ")             
//						)
//				);
//	}
}
