package net.sf.recordEditor.ui.form;

import java.awt.Component;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;

import info.clearthought.layout.TableLayout;
import info.clearthought.layout.TableLayoutConstraints;
import net.sf.recordEditor.ui.def.component.IUiComponent;

/**
 * 
 * @author Bruce Martin
 *
 */
public class TwoColumnFormBuilder {

	public static final int FIELD_HEIGHT_INDEX = 1;
	public static final int GAP_HEIGHT_INDEX = 2;

	private static final double BORDER = 10;
	private static final double[] VERTICAL_GAPS_DEFAULT 
			= {BORDER, FormHelper.PREFERRED, 5, FormHelper.FILL, BORDER};
	private static final double[] HORIZONTAL_GAPS_DEFAULT = {
			BORDER, 
			FormHelper.PREFERRED, 
			FormHelper.GAP, 5};

	private final String formName;
	private final double[] rowGaps; 

	private final JPanel pnl;
	private final TableLayout tblLayout = new TableLayout();
	private final CellPosition labelPosition, defaultFieldPosition;
	
	private int rowNum=-2;
	
	private double[] lSize = new double[40];
	
	public final int labelColumn, fieldColumnCount, lastFieldColumn;
	
	
	public TwoColumnFormBuilder(String formName, JPanel panel,
			double[] rowGaps, double[] colGaps, 
			CellPosition labelPosition, CellPosition defaultFieldPosition,
			int fieldColCount) {
		super();
		
		colGaps = colGaps == null ? VERTICAL_GAPS_DEFAULT : colGaps;
		rowGaps = rowGaps == null ? HORIZONTAL_GAPS_DEFAULT : rowGaps;
		
		this.fieldColumnCount =  (fieldColCount) * 2 - 1;
		this.lastFieldColumn = 2 + this.fieldColumnCount;
		int cols = 4 + this.fieldColumnCount;
		if (colGaps.length < cols) {
			double[] tmp = new double[cols];
			System.arraycopy(colGaps, 0, tmp, 0, colGaps.length - 1);
			for (int i = 0; i < fieldColCount-1; i++) {
				tmp[colGaps.length + i * 2 - 1 ] = colGaps[2];
				tmp[colGaps.length + i * 2  ] = FormHelper.PREFERRED;
			}
			tmp[tmp.length - 1] = colGaps[colGaps.length - 1];
			colGaps = tmp;
		}
		
		this.formName = formName + ".";
		this.rowGaps = rowGaps;
		this.labelColumn = colGaps[0] == 0 ? 0 : 1;
		this.labelPosition = labelPosition;
		this.defaultFieldPosition = defaultFieldPosition;
		
		tblLayout.setColumn(colGaps);
		
		if (rowGaps[0] > 0) {
			lSize[0] = rowGaps[0];
			rowNum+=1;
		}
			
		if (panel == null) {
			this.pnl = new JPanel(tblLayout);
		} else {
			this.pnl = panel;
			panel.setLayout(tblLayout);
		}
	}
	
	/**
	 * Create a line with one field in it
	 */
	public TwoColumnFormBuilder l(
			JComponent field, CellPosition fieldPosition, 
			int rowCount, int firstCol, int lastCol ) {
		incRow();

		return cell(field, fieldPosition, rowCount, firstCol, lastCol);
	}
	
	/**
	 * Position a cell anywhere in the line
	 */
	public TwoColumnFormBuilder cell(
			JComponent field, CellPosition fieldPosition, 
			int rowCount, int firstCol, int lastCol ) {
		
		pnl.add(field,
				new TableLayoutConstraints(firstCol, rowNum, lastCol, rowNum + rowCount-1,
						fieldPosition.horizontalPosition.cellPosition,
						fieldPosition.verticalPosition.cellPosition));
		return this;
	}

	public TwoColumnFormBuilder l(String label, IUiComponent field) {
		return l(label, field, defaultFieldPosition);
	}
	
	public TwoColumnFormBuilder l(String label, IUiComponent field, CellPosition fieldPosition) {
		JLabel jLabel = null;
		Component jComponent = field == null ? null : field.getUiContainer();
		if (label == null || label.length() == 0) {
			label = jComponent == null ? "" : jComponent.getName();
		} else {
			jLabel = new JLabel(label);
			
			if (field != null) {
				field.setFieldsLabel(jLabel);
			}
		}
		line(jLabel, label, jComponent, fieldPosition);
		
		return this;
	}
	
	
	/**
	 * Line using swing fields
	 * 
	 */
	public TwoColumnFormBuilder ls(String label, JComponent field) {
		line(label == null ? null : new JLabel(label), label, field, defaultFieldPosition);
		return this;
	}
	
	
	/**
	 * Line using swing fields
	 * 
	 */
	public TwoColumnFormBuilder ls(JComponent label, JComponent field, CellPosition fieldPosition ) {
		line(label, null, field, fieldPosition);
		return this;
	}
	
	private void line(JComponent label, String name, Component itm, CellPosition fieldPosition) {
		
		incRow();
		if (label != null) {
			pnl.add(label,
					new TableLayoutConstraints(1, rowNum, 1, rowNum,
							labelPosition.horizontalPosition.cellPosition,
							labelPosition.verticalPosition.cellPosition));
		}
		if (itm != null) {
			itm.setName(formName + name);
			pnl.add(itm,
					new TableLayoutConstraints(3, rowNum, 2 + fieldPosition.numberOfColumns(fieldColumnCount), rowNum,
							fieldPosition.horizontalPosition.cellPosition,
							fieldPosition.verticalPosition.cellPosition));
			//lSize[rowNum] = Math.max(lSize[rowNum], itm.getPreferredSize().height);
		}

	}

	/**
	 * Increment the row counter and set standard Height and Gap
	 *
	 */
	public final void incRow() {

		incRow(1, rowGaps[FIELD_HEIGHT_INDEX], rowGaps[GAP_HEIGHT_INDEX]);
	}


	/**
	 * Increment the row counter and set the Hieght and Gap amounts to the
	 * supplied values
	 *
	 * @param height Height of the Field
	 * @param gap    Gap that will follow the field
	 */
	public final void incRow(int rows, double height, double gap) {
		rowNum += 2 * rows;
		
		if (rowNum+1 >= lSize.length) {
			double[] tmp = new double[rowNum + 41];
			System.arraycopy(lSize, 0, tmp, 0, lSize.length);
			lSize = tmp;
		}

		lSize[rowNum]     = height;
		lSize[rowNum + 1] = gap;
	}
	
	/**
	 * Set the Horizontal Gap between the current Field and the Next Field
	 *
	 * @param gap Line gap
	 */
	public final TwoColumnFormBuilder setGapRE(double gap) {
		lSize[rowNum + 1] = gap;
		return this;
	}

	public JPanel build() {
		
		double[] hGap = new double[rowNum + 2];
		
		System.arraycopy(lSize, 0, hGap, 0, rowNum + 2);
		tblLayout.setRow(hGap);
		return pnl;
	}
}
